/* Sticky footer styles
-------------------------------------------------- */
html {
  position: relative;
  min-height: 100%;
}
body {
  /* Margin bottom by footer height */
  margin-bottom: 60px;
}
.footer {
  position: absolute;
  bottom: 0;
  width: 100%;
  /* Set the fixed height of the footer here */
  height: 60px;
  line-height: 60px; /* Vertically center the text there */
  background-color: #f5f5f5;
}

/* Adjustments for "How-To" on home page
-------------------------------------------------- */

.home-text img {
    width: 60%;
    margin: 1em 0 1em 20%;
    box-shadow: 12px 12px 2px 1px lightgray;
}
